"""Path module"""
from sknetwork.path.metrics import diameter
from sknetwork.path.search import breadth_first_search, depth_first_search
from sknetwork.path.shortest_path import distance, shortest_path
