"""path module"""
from sknetwork.topology.kcliques import Cliques
from sknetwork.topology.kcore import CoreDecomposition
from sknetwork.topology.triangles import Triangles

from sknetwork.topology.dag import DAG
from sknetwork.topology.structure import is_acyclic, is_bipartite, largest_connected_component, connected_components
from sknetwork.topology.weisfeiler_lehman import WeisfeilerLehman, are_isomorphic
