#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Top-level package for scikit-network"""

__author__ = """scikit-network team"""
__email__ = "bonald@enst.fr"
__version__ = '0.20.0'

import sknetwork.topology
import sknetwork.path
import sknetwork.classification
import sknetwork.clustering
import sknetwork.embedding
import sknetwork.hierarchy
import sknetwork.linalg
import sknetwork.linkpred
import sknetwork.ranking
import sknetwork.data
import sknetwork.utils
import sknetwork.visualization
